/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.aopalliance.intercept.MethodInvocation;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.aop.IntroductionInterceptor;

/**
 * Intercepts finder DAO calls and calls execute*Finder methods on dao.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class FinderIntroductionInterceptor implements IntroductionInterceptor {

  private QueryArgumentTransformer queryArgumentTransformer = new DefaultQueryArgumentTransformer();

  /** {@inheritDoc} */
  @Override
  public boolean implementsInterface(final Class intf) {
    return intf.isInterface() && FinderExecutor.class.isAssignableFrom(intf);
  }

  /** {@inheritDoc} */
  @Override
  public Object invoke(@NotNull final MethodInvocation invocation) throws Throwable {
    @NotNull
    final FinderExecutor dao = (FinderExecutor) invocation.getThis();
    @NotNull
    final Method meth = invocation.getMethod();
    if (Utils.isFinder(meth)) {
      @NotNull
      final Object[] arguments = invocation.getArguments();
      if (Utils.isSingleFind(invocation.getMethod())) {
        @NotNull
        final Map<String, ?> namedArgs = getNamedParams(meth, arguments, Integer.MAX_VALUE, Integer.MAX_VALUE);
        if (namedArgs.isEmpty()) {
          return dao.executeSingleFinder(meth, arguments);
        } else {
          return dao.executeSingleFinder(meth, namedArgs);
        }
      } else {
        @Nullable
        final Integer limitIndex = Utils.getLimitParameterIndex(meth);
        @Nullable
        final Integer offsetIndex = Utils.getOffsetParameterIndex(meth);
        @Nullable
        final Integer limit = limitIndex == null ? null : (Integer) arguments[limitIndex];
        @Nullable
        final Integer offset = offsetIndex == null ? null : (Integer) arguments[offsetIndex];
        final int queryArgsNum = arguments.length
            - (limitIndex == null ? 0 : 1)
            - (offsetIndex != null && !offsetIndex.equals(limitIndex) ? 1 : 0);
        @NotNull
        final Object[] queryArgs = new Object[queryArgsNum];
        int j = 0;
        for (int i = 0; i < arguments.length; i++) {
          if (limitIndex != null && i == limitIndex || offsetIndex != null && i == offsetIndex) {
            continue;
          }
          queryArgs[j] = queryArgumentTransformer.transform(meth.getParameterTypes()[i], meth.getParameterAnnotations()[i], arguments[i]);
          j++;
        }
        @NotNull
        final Map<String, ?> namedArgs = getNamedParams(meth, queryArgs, limitIndex, offsetIndex);
        final List<Object> data;
        if (namedArgs.isEmpty()) {
          data = dao.executeListFinder(meth, queryArgs, limit, offset);
        } else {
          data = dao.executeListFinder(meth, namedArgs, limit, offset);
        }
        final Collection<Object> result;
        final Finder finder = meth.getAnnotation(Finder.class);
        if (finder != null && finder.returnAs() != List.class) {
          @SuppressWarnings({ "unchecked" })
          final Collection<Object> c = finder.returnAs().newInstance();
          c.addAll(data);
          result = c;
        } else {
          result = data;
        }
        return result;
      }
    } else {
      return invocation.proceed();
    }
  }

  @NotNull
  private static Map<String, ?> getNamedParams(@NotNull final Method meth,
                                               @NotNull final Object[] queryArgs,
                                               @Nullable final Integer limitIndex,
                                               @Nullable final Integer offsetIndex) {
    final Map<String, Object> result = new HashMap<String, Object>();
    if (queryArgs.length > 0) {
      for (int i = 0; i < queryArgs.length; i++) {
        final int paramIndex = i
            + (limitIndex != null && i >= limitIndex ? 1 : 0)
            + (offsetIndex != null && i >= offsetIndex && !offsetIndex.equals(limitIndex) ? 1 : 0);
        Named n = Utils.getNamed(meth.getParameterAnnotations()[paramIndex]);
        if (n == null) {
          return Collections.emptyMap();
        } else {
          result.put(n.value(), queryArgs[i]);
        }
      }
    }
    return result;
  }

  public void setQueryArgumentTransformer(final QueryArgumentTransformer queryArgumentTransformer) {
    this.queryArgumentTransformer = queryArgumentTransformer;
  }
}
