/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import net.sf.autodao.Dao;
import net.sf.autodao.Finder;
import net.sf.autodao.FinderExecutor;
import net.sf.autodao.PersistentEntity;
import net.sf.autodao.Utils;
import org.hibernate.Query;
import org.hibernate.Session;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.beans.factory.BeanNameAware;
import org.springframework.beans.factory.annotation.Required;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.support.HibernateDaoSupport;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import org.springframework.util.StringUtils;

/**
 * Hibernate implementation of AutoDAO. This class is intentionally made final. If you need custom DAO method, introduce
 * additional bean.
 * <p/>
 * NOT FOR PUBLIC USE.
 * @param <E>
 * @param <K>
 */
@Transactional(propagation = Propagation.MANDATORY)
public final class HibernateDaoImpl<E extends PersistentEntity<K>, K extends Serializable> extends HibernateDaoSupport
    implements Dao<E, K>, FinderExecutor, BeanNameAware {

  /** Bean name. */
  @NotNull
  private String beanName;

  /** class of persistent object. */
  @NotNull
  private Class<E> type;

  /** {@inheritDoc} */
  @Override
  public void create(@NotNull final E o) {
    this.checkObject(o);
    Assert.isTrue(!this.getHibernateTemplate().contains(o), "You're trying to persist already persistent entity: " + o);
    this.getHibernateTemplate().persist(o);
  }

  /** {@inheritDoc} */
  @Override
  public void delete(@NotNull final E o) {
    this.checkObject(o);
    Assert.isTrue(this.getHibernateTemplate().contains(o), "You're trying to delete transient/detached entity: " + o);
    this.getHibernateTemplate().delete(o);
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public <T> List<T> executeListFinder(@NotNull final Method method,
                                       @NotNull final Map<String, ?> queryArgs,
                                       @Nullable final Integer limit,
                                       @Nullable final Integer offset) {
    return this.getHibernateTemplate().executeFind(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, limit, offset), queryArgs).list();
      }
    });
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public <T> List<T> executeListFinder(@NotNull final Method method,
                                       @NotNull final Object[] queryArgs,
                                       @Nullable final Integer limit,
                                       @Nullable final Integer offset) {
    return this.getHibernateTemplate().executeFind(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, limit, offset), queryArgs).list();
      }
    });
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Map<String, ?> queryArgs) {
    return (T) this.getHibernateTemplate().execute(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, null, null), queryArgs).uniqueResult();
      }
    });
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Object[] queryArgs) {
    return (T) this.getHibernateTemplate().execute(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, null, null), queryArgs).uniqueResult();
      }
    });
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public E get(@NotNull final K id) {
    Assert.notNull(id, "Id cannot be null");
    return (E) this.getHibernateTemplate().get(this.type, id);
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public E merge(@NotNull final E o) {
    return (E) this.getHibernateTemplate().merge(o);
  }

  /** {@inheritDoc} */
  @Override
  public void refresh(@NotNull final E o) {
    this.getHibernateTemplate().refresh(o);
  }

  /** {@inheritDoc} */
  @Override
  public void saveOrUpdate(@NotNull final E o) {
    this.checkObject(o);
    this.getHibernateTemplate().saveOrUpdate(o);
  }

  /** {@inheritDoc} */
  @Override
  public void setBeanName(@NotNull final String name) {
    this.beanName = name;
  }

  @Required
  public void setType(@NotNull final Class<E> type) {
    this.type = type;
  }

  /** {@inheritDoc} */
  @Override
  public String toString() {
    return "AutoDao [" + this.type.getName() + ']';
  }

  /** {@inheritDoc} */
  @Override
  public void update(@NotNull final E o) {
    this.checkObject(o);
    Assert.isTrue(this.getHibernateTemplate().contains(o), "You're trying to update transient/detached entity: " + o);
    this.getHibernateTemplate().update(o);
  }

  /** {@inheritDoc} */
  @Override
  protected void initDao() {
    Assert.notNull(this.type, "You forgot to call #setType() on  bean '" + this.beanName + '\'');
  }

  @NotNull
  private Query buildQuery(@NotNull final Session s,
                           @NotNull final Method m,
                           @Nullable final Integer limit,
                           @Nullable final Integer offset) {
    @Nullable
    final Finder f = m.getAnnotation(Finder.class);
    @NotNull
    final Query q;
    if (f != null && StringUtils.hasText(f.query())) {
      q = s.createQuery(f.query());
    } else {
      final String queryName;
      if (f != null && StringUtils.hasText(f.queryName())) {
        queryName = f.queryName();
      } else {
        queryName = HibernateUtils.getNamedQueryName(s.getSessionFactory(), this.type, m);
      }
      q = s.getNamedQuery(queryName);
    }
    if (limit != null) {
      q.setMaxResults(limit);
    }
    if (offset != null) {
      q.setFirstResult(offset);
    }
    return q;
  }

  /**
   * Check entity.
   * @param o entity.
   */
  private void checkObject(@NotNull final E o) {
    Assert.notNull(o, "Entity cannot be null");
    Assert.isAssignable(this.type, o.getClass(), "Invalid entity type");
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Map<String, ?> params) {
    for (final Map.Entry<String, ?> o : params.entrySet()) {
      @NotNull
      final String paramName = o.getKey();
      @Nullable
      final Object value = o.getValue();
      if (value == null) {
        q.setParameter(paramName, null);
      } else if (value.getClass().isArray()) {
        q.setParameterList(paramName, Utils.toObjectArray(value));
      } else if (Collection.class.isAssignableFrom(value.getClass())) {
        q.setParameterList(paramName, (Collection) value);
      } else {
        q.setParameter(paramName, value);
      }
    }
    return q;
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Object[] params) {
    int i = 0;
    for (final Object o : params) {
      q.setParameter(i, o);
      i++;
    }
    return q;
  }
}
